#!/usr/bin/env python
# Python 3.*

from gcodeWriter import *
from termcolor import colored
import math

class grid(gcodeTerminalGenerator):
    def __init__(self, parameters):
        self.computedParameters(parameters)
        gcodeTerminalGenerator.__init__(self)
    
    def computedParameters(self, parameters):
        self.parameters = parameters
        if (self.parameters['segment']['max'] == 'auto'):
            self.parameters['segment']['max'] = math.floor(self.parameters['dimensions']['width'] / (self.parameters['nozzle'] * self.parameters['ratio']['width']))
        self.parameters['layerNumber'] = math.floor(self.parameters['dimensions']['height'] / (self.parameters['nozzle'] * self.parameters['ratio']['height']))
        self.parameters['startPosition'] = {
                'X': self.parameters['center']['X'] - self.parameters['dimensions']['width'] / 2 + self.parameters['nozzle'] * self.parameters['ratio']['width'] / 2, 
                'Y': self.parameters['center']['Y'] - self.parameters['dimensions']['width'] / 2 + self.parameters['nozzle'] * self.parameters['ratio']['width'] / 2
        }
        self.parameters['endPosition'] = {
                'X': self.parameters['startPosition']['X'] + self.parameters['dimensions']['width'] - self.parameters['nozzle'] * self.parameters['ratio']['width'], 
                'Y': self.parameters['startPosition']['Y'] + self.parameters['dimensions']['width'] - self.parameters['nozzle'] * self.parameters['ratio']['width']
        }
        self.parameters['ratio']['extruder'] = round(self.parameters['ratio']['width'] * self.parameters['ratio']['height'] * self.parameters['nozzle'] ** 2 / (self.parameters['filament'] ** 2), 5)
           
    def app(self):
        self.layers = []
        self.generateLayers()
        self.writeInfo()
        self.writeGcode()    
        
    def writeInfo(self):
        self.writer.addComment('3D grid object')   
        for key, value in self.parameters.items():
            self.writer.addComment('%s: %s'%(key, value))     
        return self
    
    def writeGcode(self):
        self.writer.setUnitsToMillimeters()
        self.writer.fanOff()
        self.writer.useAbsoluteCoordinates()
        self.writer.resetExtruder()
        self.writer.useAbsoluteDistances()
        for layer in self.layers:
             layer.writeGcode()
        self.writer.setTemperature(0)
        self.writer.fanOff()
        return self
        
    def addLayer(self, index):
        self.layers.append(layer(self.parameters, self.writer, index))
        return layer
        
    def generateLayers(self):
        for index in range(1, self.parameters['layerNumber'] + 1):
            self.writer.coordinates['E'] = 0
            self.addLayer(index)
        return self.layers


class layer:
    def __init__(self, parameters, writer, index):
        self.parameters = parameters
        self.writer = writer
        self.index = index
        self.segmentNumber = self.repartitionFunction(self.index, 1, self.parameters['segment']['max'], self.parameters['layerNumber'], self.parameters['segment']['min'])
        self.segmentSpacing = self.parameters['dimensions']['width'] / (self.segmentNumber + 1)
        self.segments = []
        self.generateSegments()
    
    def writeGcode(self):
        self.writer.resetExtruder()
        if (self.index == 3):
            self.writer.fanOn()
        for segment in self.segments:
            segment.writeGcode()
        return self
            
    def addSegment(self, coordinates):
        self.segments.append(segment(self.parameters, self.writer, coordinates, self.index == 1))
        return segment
    
    def generateSegments(self):
        if (self.index == 1):
            self.addSegment({
                'Z': self.index * self.parameters['nozzle'] * self.parameters['ratio']['height'], 
                'F': 720        
            })
        elif (self.index == 2):
            self.addSegment({
                'Z': self.index * self.parameters['nozzle'] * self.parameters['ratio']['height'], 
                'F': 1400        
            })
        else:
            self.addSegment({
                'Z': self.index * self.parameters['nozzle'] * self.parameters['ratio']['height']         
            })
        if (self.index % 2 == 0):
            self.generatePerimetersEven()
            self.generateInfillEven()
        else:
            self.generatePerimetersOdd()
            self.generateInfillOdd()  
        return self.segments
    
    def generatePerimetersEven(self):
        for perimeterIndex in range(0, self.parameters['perimeters']):
            offset = perimeterIndex * self.parameters['nozzle'] * self.parameters['ratio']['width']
            self.addSegment({
                'X': self.parameters['endPosition']['X'] - offset, 
                'Y': self.parameters['endPosition']['Y'] - offset       
            })
            self.addSegment({
                'X': self.parameters['startPosition']['X'] + offset, 
                'Y': self.parameters['endPosition']['Y'] - offset,
                'E': 'XY'        
            })
            self.addSegment({
                'X': self.parameters['startPosition']['X'] + offset, 
                'Y': self.parameters['startPosition']['Y'] + offset,
                'E': 'XY'        
            })
            self.addSegment({
                'X': self.parameters['endPosition']['X'] - offset, 
                'Y': self.parameters['startPosition']['Y'] + offset,
                'E': 'XY'        
            })
            self.addSegment({
                'X': self.parameters['endPosition']['X'] - offset, 
                'Y': self.parameters['endPosition']['Y'] - offset,
                'E': 'XY'         
            })
            
    def generatePerimetersOdd(self):
        for perimeterIndex in range(0, self.parameters['perimeters']):
            offset = perimeterIndex * self.parameters['nozzle'] * self.parameters['ratio']['width']
            self.addSegment({
                'X': self.parameters['startPosition']['X'] + offset, 
                'Y': self.parameters['startPosition']['Y'] + offset       
            })
            self.addSegment({
                'X': self.parameters['endPosition']['X'] - offset, 
                'Y': self.parameters['startPosition']['Y'] + offset,
                'E': 'XY'        
            })
            self.addSegment({
                'X': self.parameters['endPosition']['X'] - offset, 
                'Y': self.parameters['endPosition']['Y'] - offset,
                'E': 'XY'        
            })
            self.addSegment({
                'X': self.parameters['startPosition']['X'] + offset, 
                'Y': self.parameters['endPosition']['Y'] - offset,
                'E': 'XY'        
            })
            self.addSegment({
                'X': self.parameters['startPosition']['X'] + offset, 
                'Y': self.parameters['startPosition']['Y'] + offset,
                'E': 'XY'         
            })
            
    def generateInfillEven(self):
        for infillIndex in range(1, self.segmentNumber + 1):
            offset = infillIndex * self.segmentSpacing
            if (infillIndex % 2 == 0):
                self.addSegment({
                    'X': self.parameters['endPosition']['X'] - offset, 
                    'Y': self.parameters['startPosition']['Y']         
                })
                self.addSegment({
                    'X': self.parameters['endPosition']['X'] - offset, 
                    'Y': self.parameters['endPosition']['Y'],
                    'E': 'XY'         
                })
            else:                
                self.addSegment({
                    'X': self.parameters['endPosition']['X'] - offset, 
                    'Y': self.parameters['endPosition']['Y']         
                })
                self.addSegment({
                    'X': self.parameters['endPosition']['X'] - offset, 
                    'Y': self.parameters['startPosition']['Y'],
                    'E': 'XY'         
                })
            
    def generateInfillOdd(self):
        for infillIndex in range(1, self.segmentNumber + 1):
            offset = infillIndex * self.segmentSpacing
            if (infillIndex % 2 == 0):
                self.addSegment({
                    'X': self.parameters['endPosition']['X'], 
                    'Y': self.parameters['startPosition']['Y'] + offset
                })
                self.addSegment({
                    'X': self.parameters['startPosition']['X'], 
                    'Y': self.parameters['startPosition']['Y'] + offset,
                    'E': 'XY'
                })  
            else:
                self.addSegment({
                    'X': self.parameters['startPosition']['X'], 
                    'Y': self.parameters['startPosition']['Y'] + offset       
                })
                self.addSegment({
                    'X': self.parameters['endPosition']['X'], 
                    'Y': self.parameters['startPosition']['Y'] + offset,
                    'E': 'XY'         
                })
                
    def repartitionFunction(self, x, x0, y0, x1, y1):
        return math.floor((y1 - y0) / (x1 - x0) * (x - x0) + y0)


class segment:
    def __init__(self, parameters, writer, coordinates, firstLayer = False):
        self.parameters = parameters
        self.writer = writer
        self.firstLayer = firstLayer
        self.coordinates = coordinates
        self.retraction = False
        if ('X' in self.coordinates.keys()): 
            XDelta = self.coordinates['X'] - self.writer.coordinates['X']
        else:
            XDelta = 0
        if ('Y' in self.coordinates.keys()): 
            YDelta = self.coordinates['Y'] - self.writer.coordinates['Y']
        else:
            YDelta = 0
        travel = math.sqrt(XDelta ** 2 + YDelta ** 2)
        
        if ('E' in self.coordinates.keys() and self.coordinates['E'] == 'XY'):
            if (self.firstLayer):
                self.coordinates['E'] = self.writer.coordinates['E'] + travel * self.parameters['ratio']['extruder'] * self.parameters['ratio']['firstLayer']
            else:
                self.coordinates['E'] = self.writer.coordinates['E'] + travel * self.parameters['ratio']['extruder']
        elif (travel > self.parameters['retraction']['minTravel']):
            self.retraction = True
            self.localE = self.writer.coordinates['E']
        for key in self.coordinates:
            self.writer.coordinates[key] = self.coordinates[key]

    def writeGcode(self):
        if (self.retraction):
            self.writer.retractedMove(self.coordinates, self.parameters['retraction'], self.localE)
        else:
            self.writer.move(self.coordinates)
        
      
grid({
    'nozzle'                   : 0.35,
    'filament'                 : 3,
    'dimensions': {'width'     : 52, 
                   'height'    : 68},
    'segment':    {'min'       : 1, 
                   'max'       : 25},
    'ratio':      {'width'     : 1,
                   'height'    : 0.8,
                   'firstLayer': 1.2},
    'perimeters'               : 2,
    'center':     {'X'         : 100,
                   'Y'         : 100},
    'retraction': {'length'    : 1.4,
                   'minTravel' : 5}
})
